CREATE TABLE Schema (
  boolean BOOLEAN,
  int8 INT8,
  int16 INT16,
  int32 INT32,
  int64 INT,
  uint8 UINT8,
  text TEXT,
  bytea BYTEA,
  inet INET,
  date DATE,
  timestamp TIMESTAMP,
  time TIME,
  interval INTERVAL,
  uuid UUID,
  map MAP,
  list LIST
);
